#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2020-02-24
# @Filename: base.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

import abc
import datetime
import inspect
import re
import warnings

import numpy
import peewee
from astropy import table

from sdssdb.peewee import BaseModel
from sdssdb.peewee.sdss5db import catalogdb as cdb
from sdssdb.peewee.sdss5db import targetdb as tdb
from sdsstools import read_yaml_file
from sdsstools._vendor.color_print import color_text

from target_selection import __version__, config, log
from target_selection.exceptions import TargetSelectionError
from target_selection.utils import Timer


EPOCH = 2016.0


class BaseCarton(metaclass=abc.ABCMeta):
    """A base class for target cartons.

    This class is not intended for direct instantiation. Instead, it must be
    subclassed and the relevant class attributes overridden with the values
    corresponding to the carton.

    Parameters
    ----------
    targeting_plan : str
        The target selection plan version.
    config_file : str
        The path to the configuration file to use. If undefined, uses the
        internal ``target_selection.yml`` file.
    schema : str
        Schema in which the temporary table with the results of the
        query will be created. If `None`, tries to use the ``schema`` parameter
        from the configuration file for this plan of target selection. If
        the parameter is not set, defaults to ``'sandbox'``.
    table_name : str
        The name of the temporary table. Defaults to ``temp_<name>`` where
        ``<name>`` is the target class name.

    Attributes
    ----------
    name : str
        The name of the carton (required).
    cadence : str
        The label of the cadence rule for this carton.
    category : str
        The category of targets for this carton.
    mapper : str
        The mapper with which this carton is associated.
    orm : str
        The ORM library to be used, ``peewee`` or ``sqlalchemy``.
    tag : str
        The version of the ``target_selection`` code used.
    query_region : tuple
        A tuple defining the region over which the query should be performed,
        with the format ``(ra, dec, radius)`` in degrees. This will append a
        ``q3c_radial_query`` condition to the query.
    load_magnitudes : bool
        Whether to load target magnitudes. In general this must be `True`
        except for cartons for which it's known the magnitudes will not be
        used, e.g., skies.

    """

    name = None
    cadence = None
    category = None
    program = None
    mapper = None
    priority = None
    value = None
    instrument = None
    can_offset = False

    load_magnitudes = True

    query_region = None

    def __init__(self, targeting_plan, config_file=None, schema=None, table_name=None):
        assert self.name, "carton subclass must override name"
        assert self.category, "carton subclass must override category"
        assert self.program, "carton subclass must override program"

        self.plan = targeting_plan
        self.tag = __version__ if self.plan != "0.0.0-test" else None

        if config_file:
            this_config = read_yaml_file(config_file)
        else:
            this_config = config.copy()

        if self.plan not in this_config:
            raise TargetSelectionError(f"({self.name}): cannot find plan {self.plan!r} in config.")

        self.config = this_config[self.plan]

        if "parameters" in self.config:
            self.parameters = self.config["parameters"].get(self.name, None)
        else:
            self.parameters = None

        try:
            self.xmatch_plan = self.config["xmatch_plan"]
        except KeyError:
            raise TargetSelectionError(f"({self.name}): xmatch_plan not found in config.")

        # Check the signature of build_query
        self._build_query_signature = inspect.signature(self.build_query)
        if "version_id" not in self._build_query_signature.parameters:
            raise TargetSelectionError("build_query does not accept version_id")

        self.database = tdb.database
        assert self.database.connected, "database is not connected."

        self.schema = schema or self.config.get("schema", None) or "sandbox"
        self.table_name = table_name or f"temp_{self.name}".replace("-", "_")
        self.RModel = None

        if self.cadence:
            ncad = tdb.Cadence.select().where(tdb.Cadence.label == self.cadence).count()
            assert ncad == 1, f"{self.cadence!r} does not exist in targetdb.cadence."

        self.log = log
        self.has_run = False
        self._disable_query_log = False

        # We cannot set temp_buffers multiple times if there are temporary tables
        # (as in add_optical_magnitudes) so we set it here.
        if "database_options" in self.config and "temp_buffers" in self.config["database_options"]:
            temp_buffers = self.config["database_options"].pop("temp_buffers")
            self.database.execute_sql(f"SET temp_buffers = '{temp_buffers}'")

    @property
    def path(self):
        """The schema-qualified path to the output table."""

        return self.schema + "." + self.table_name

    @abc.abstractmethod
    def build_query(self, version_id, query_region=None):
        """Builds and returns the query.

        The ORM query for the target class. Note that this must be the
        *un-executed* query, which will be executed in `.run`. The select
        statement must be run on ``catalogdb.catalog`` and needs to include,
        at least, ``catalogid``. Additional columns such as
        ``ra, dec, pmra, pmdec`` can be specified but are otherwise completed
        when calling `.run`. Magnitude columns can be included and propagated
        to ``targetdb`` but must be aliased as ``magnitude_<band>`` where
        ``<band>`` must be one of the columns in ``targetdb.magnitude``.

        The query returned must include a filter for ``version_id`` on
        ``catalogdb.catalog``. Normally this is applied to the ``Catalog``
        select as ``.where(Catalog.version_id == version_id)``.

        Parameters
        ----------
        version_id : int
            The id of the cross-match version to use.

        Returns
        -------
        query
            A :class:`peewee:Select` or :class:`peewee:ModelSelect` query.

        """

        pass

    def get_model(self):
        """Returns a Peewee model for the temporary table using reflection."""

        # peewee has a Model class, BaseModel class, and ModelBase class.
        # The below Model class  (after this comment)
        # is different from peewee Model class.
        # The below BaseModel class (after this comment) is from the
        # sdssdb.peewee package's __init__.py.
        # It is different from the peewee BaseModel class.
        #
        # The below line is from sdssdb/peewee/__init__.py
        # This Model class in the below line is the peewee Model class.
        #     class BaseModel(Model, metaclass=ReflectMeta):
        #
        # The above line, relates the below Model class to
        # the ReflectMeta metaclass via the BaseModel class.
        # Inside the ReflectMeta metaclass is the code in __new__()
        # which generates the Model class.
        #
        # So defining the Model class here leads to creation of
        # the target_selection Model class. Later below we have the statement
        # "return Model" which is the last statement of get_model().
        #
        class Model(BaseModel):
            catalogid = peewee.BigIntegerField(primary_key=True)
            selected = peewee.BooleanField()

            class Meta:
                database = self.database
                table_name = self.table_name
                schema = self.schema
                reflection_options = {
                    "skip_foreign_keys": True,
                    "use_peewee_reflection": False,
                    "force": True,
                }
                use_reflection = True

        if not Model.table_exists():
            raise TargetSelectionError(f"temporary table {self.path!r} does not exist.")

        return Model

    def run(
        self,
        query_region=None,
        overwrite=False,
        limit=None,
        add_optical_magnitudes=True,
        **post_process_kawrgs,
    ):
        """Executes the query and post-process steps, and stores the results.

        This method calls `.build_query` and runs the returned query. The
        output of the query is stored in a temporary table whose schema and
        table name are defined when the object is instantiated.

        After the query has run, the `.post_process` routine is called if the
        method has been overridden for the given carton.

        Parameters
        ----------
        query_region : tuple
            A tuple defining the region over which the query should be
            performed, with the format ``(ra, dec, radius)`` in degrees. This
            will append a ``q3c_radial_query`` condition to the query.
        overwrite : bool
            Whether to overwrite the intermediary table if already exists.
        limit : int or None
            Limit the query to this number of targets. Useful for testing.
            The ``LIMIT`` statement is added to the query returned by
            `.build_query` and the exact behaviour will depend on the query.
        post_process_args : dict
            Keyword arguments to be passed to `.post_process`.

        Returns
        -------
        model : :class:`peewee:Model`
            The model for the intermediate table.

        """

        query_region = query_region or self.query_region

        path = self.path
        execute_sql = self.database.execute_sql

        if self.database.table_exists(self.table_name, schema=self.schema):
            if overwrite:
                self.log.info(f"Dropping table {path!r}.")
                self.drop_table()
            else:
                raise RuntimeError(f"Temporary table {path!r} already exists.")

        self.log.info("Running query ...")
        version_id = self.get_version_id()

        with Timer() as timer:
            # If build_query accepts a query_region parameter, call with the query region.
            # Otherwise will add the radial query condition later.
            if "query_region" in self._build_query_signature.parameters:
                query = self.build_query(version_id, query_region=query_region)
            else:
                query = self.build_query(version_id)

            # Make sure the catalogid column is selected.
            if cdb.Catalog.catalogid not in query._returning:
                raise RuntimeError("catalogid is not being returned in query.")

            if query_region:
                if "query_region" in self._build_query_signature.parameters:
                    pass
                else:
                    # This may be quite inefficient depending on the query.
                    subq = query.alias("subq")
                    query = (
                        peewee.Select(columns=[peewee.SQL("subq.*")])
                        .from_(subq)
                        .join(cdb.Catalog, on=(cdb.Catalog.catalogid == subq.c.catalogid))
                        .where(
                            peewee.fn.q3c_radial_query(
                                cdb.Catalog.ra,
                                cdb.Catalog.dec,
                                query_region[0],
                                query_region[1],
                                query_region[2],
                            )
                        )
                    )

            if limit:
                query = query.limit(limit)

            query_sql, params = self.database.get_sql_context().sql(query).query()
            cursor = self.database.cursor()
            query_str = cursor.mogrify(query_sql, params).decode()

            if not self._disable_query_log:
                log_message = f"CREATE TABLE IF NOT EXISTS {path} AS " + query_str
                if self.log.rich_console:
                    self.log.debug(log_message, extra={"highlighter": None})
                else:
                    self.log.debug(color_text(log_message, "darkgrey"))
            else:
                self.log.debug("Not printing VERY long query.")

            with self.database.atomic():
                self.setup_transaction()
                execute_sql(f"CREATE TABLE IF NOT EXISTS {path} AS " + query_sql, params)

        self.log.info(f"Created table {path!r} in {timer.interval:.3f} s.")

        # Note that above execute_sql() is the same as below
        # self.database.execute_sql() since at the top of this method
        # we have the below statement.
        #        execute_sql = self.database.execute_sql

        self.database.execute_sql("COMMIT;")
        self.RModel = self.get_model()

        self.log.debug("Adding columns and indexes.")

        columns = [col.name for col in self.database.get_columns(self.table_name, self.schema)]

        if "selected" not in columns:
            execute_sql(f"ALTER TABLE {path} ADD COLUMN selected BOOL DEFAULT TRUE;")
            self.RModel._meta.add_field("selected", peewee.BooleanField())

        if "cadence" not in columns and self.cadence is None:
            execute_sql(f"ALTER TABLE {path} ADD COLUMN cadence VARCHAR;")
            self.RModel._meta.add_field("cadence", peewee.TextField())

        if "priority" not in columns and self.priority is None:
            execute_sql(f"ALTER TABLE {path} ADD COLUMN priority INTEGER;")
            self.RModel._meta.add_field("priority", peewee.IntegerField())

        if "instrument" not in columns and self.instrument is None:
            execute_sql(f"ALTER TABLE {path} ADD COLUMN instrument TEXT;")
            self.RModel._meta.add_field("instrument", peewee.TextField())

        execute_sql(f"ALTER TABLE {path} ADD PRIMARY KEY (catalogid);")
        execute_sql(f"CREATE INDEX ON {path} (selected);")
        execute_sql(f"ANALYZE {path};")

        n_rows = self.RModel.select().count()
        self.log.debug(f"Table {path!r} contains {n_rows:,} rows.")

        self.log.debug("Running post-process.")
        with self.database.atomic():
            self.setup_transaction()
            self.post_process(self.RModel, **post_process_kawrgs)

        n_selected = self.RModel.select().where(self.RModel.selected >> True).count()
        self.log.debug(f"Selected {n_selected:,} rows after post-processing.")

        if add_optical_magnitudes:
            self.log.debug("Adding optical magnitude columns.")
            self.add_optical_magnitudes()

        self.has_run = True

        return self.RModel

    def get_version_id(self):
        """Returns the version_id for the cross-match plan."""

        return cdb.Version.get(plan=self.xmatch_plan).id

    def add_optical_magnitudes(self):
        """Adds ``gri`` magnitude columns."""

        Model = self.RModel

        magnitudes = ["g", "r", "i", "z"]

        # Check if ALL the columns have already been created in the query.
        # If so, just return.
        if any([mag in Model._meta.columns for mag in magnitudes]):
            if not all([mag in Model._meta.columns for mag in magnitudes]):
                raise TargetSelectionError(
                    "Some optical magnitudes are defined in the query but not all of them."
                )
            if "optical_prov" not in Model._meta.columns:
                raise TargetSelectionError("optical_prov column does not exist.")
            self.log.warning("All optical magnitude columns are defined in the query.")
            return

        # First create the columns. Also create z to speed things up. We won't
        # use transformations for z but we can use the initial query to populate
        # it and avoid doing the same query later when loading the magnitudes.
        for mag in magnitudes:
            self.database.execute_sql(f"ALTER TABLE {self.path} ADD COLUMN {mag} REAL;")
            Model._meta.add_field(mag, peewee.FloatField())

        self.database.execute_sql(f"ALTER TABLE {self.path} ADD COLUMN optical_prov TEXT;")
        Model._meta.add_field("optical_prov", peewee.TextField())

        # Step 1: join with gaia_dr3_synthetic_photometry_gspc and use synthetic SDSS mags.

        SynPhot = cdb.Gaia_dr3_synthetic_photometry_gspc
        with self.database.atomic():
            self.database.execute_sql("DROP TABLE IF EXISTS " + self.table_name + "_g3xp")
            temp_table = peewee.Table(self.table_name + "_g3xp")

            (
                Model.select(
                    Model.catalogid,
                    SynPhot.g_sdss_mag,
                    SynPhot.r_sdss_mag,
                    SynPhot.i_sdss_mag,
                    SynPhot.z_sdss_mag,
                )
                .join(
                    cdb.CatalogToGaia_DR3,
                    on=(cdb.CatalogToGaia_DR3.catalogid == Model.catalogid),
                )
                .join(SynPhot, on=(cdb.CatalogToGaia_DR3.target_id == SynPhot.source_id))
                .where(Model.g.is_null() | Model.r.is_null() | Model.i.is_null())
                .where(Model.selected >> True)
                .where(
                    cdb.CatalogToGaia_DR3.best >> True,
                    cdb.CatalogToGaia_DR3.version_id == self.get_version_id(),
                )
                .where(
                    SynPhot.g_sdss_mag.is_null(False),
                    SynPhot.r_sdss_mag.is_null(False),
                    SynPhot.i_sdss_mag.is_null(False),
                    # Now removing the flag selection because it killed everything
                    # with r_sdss_mag < 14.2
                    # SynPhot.g_sdss_flag == 1,
                    # SynPhot.r_sdss_flag == 1,
                    # SynPhot.i_sdss_flag == 1,
                    # It's possible that in future, we might want to apply cuts on bp-rp
                    # or on the derived sdss colours
                    # Fairly arbitrary faint cut
                    SynPhot.r_sdss_mag < 15.0,
                    # Could potentially go fainter, or could cut on G,BP,RP etc instead
                    # but that would require a join to gaia_dr3_source table
                    #
                    # The following cut on SNR is completely unnecessary at G<15
                    # (SynPhot.r_sdss_flux_error <
                    #  SynPhot.r_sdss_flux / 30.0),
                )
                .create_table(temp_table._path[0], temporary=True)
            )

            nrows = (
                Model.update(
                    {
                        Model.g: temp_table.c.g_sdss_mag,
                        Model.r: temp_table.c.r_sdss_mag,
                        Model.i: temp_table.c.i_sdss_mag,
                        Model.optical_prov: peewee.Value("sdss_psfmag_from_g3xp"),
                    }
                )
                .from_(temp_table)
                .where(Model.catalogid == temp_table.c.catalogid)
                .where(
                    temp_table.c.g_sdss_mag.is_null(False)
                    & temp_table.c.r_sdss_mag.is_null(False)
                    & temp_table.c.i_sdss_mag.is_null(False)
                )
            ).execute()

        self.log.debug(f"{nrows:,} associated with Gaia DR3 XP magnitudes.")

        # Step 2: localise entries with empty magnitudes,
        # join with sdss_dr13_photoobj and use SDSS magnitudes.

        # https://www.sdss4.org/dr16/algorithms/photo_flags/
        # https://www.sdss4.org/dr16/algorithms/photo_flags_recommend/
        # we ignore SDSS photometry when any of these bits are set
        # TODO improve/expand the list of rejected bits
        sdss_quality_bitmask = (
            0  # just a placeholder
            + (1 << 7)  # NOPROFILE
            + (1 << 18)  # SATURATED
            + (1 << 22)  # BADSKY
            + (1 << 24)  # TOOLARGE
            + (1 << (16 + 32))  # TOO_FEW_GOOD_DETECTIONS
        )

        with self.database.atomic():
            self.database.execute_sql("DROP TABLE IF EXISTS " + self.table_name + "_sdss")
            temp_table = peewee.Table(self.table_name + "_sdss")

            SDSS_DR13 = cdb.SDSS_DR13_PhotoObj
            (
                Model.select(
                    Model.catalogid,
                    SDSS_DR13.psfmag_g,
                    SDSS_DR13.psfmag_r,
                    SDSS_DR13.psfmag_i,
                    SDSS_DR13.psfmag_z,
                )
                .join(
                    cdb.CatalogToSDSS_DR13_PhotoObj_Primary,
                    on=(cdb.CatalogToSDSS_DR13_PhotoObj_Primary.catalogid == Model.catalogid),
                )
                .join(
                    SDSS_DR13,
                    on=(cdb.CatalogToSDSS_DR13_PhotoObj_Primary.target_id == SDSS_DR13.objid),
                )
                .where(Model.g.is_null() | Model.r.is_null() | Model.i.is_null())
                .where(
                    cdb.CatalogToSDSS_DR13_PhotoObj_Primary.best >> True,
                    cdb.CatalogToSDSS_DR13_PhotoObj_Primary.version_id == self.get_version_id(),
                )
                .where(
                    SDSS_DR13.psfmag_r > 14.0,  # avoid stars close to saturation
                    SDSS_DR13.psfmag_g > -99.0,  # avoid bad mags
                    SDSS_DR13.psfmag_i > -99.0,  # avoid bad mags
                    SDSS_DR13.flags.bin_and(sdss_quality_bitmask) == 0,
                )
                .where(Model.selected >> True)
                .create_table(temp_table._path[0], temporary=True)
            )

            nrows = (
                Model.update(
                    {
                        Model.g: temp_table.c.psfmag_g,
                        Model.r: temp_table.c.psfmag_r,
                        Model.i: temp_table.c.psfmag_i,
                        Model.optical_prov: peewee.Value("sdss_psfmag"),
                    }
                )
                .from_(temp_table)
                .where(Model.catalogid == temp_table.c.catalogid)
                .where(
                    temp_table.c.psfmag_g.is_null(False)
                    & temp_table.c.psfmag_r.is_null(False)
                    & temp_table.c.psfmag_i.is_null(False)
                )
            ).execute()

        self.log.debug(f"{nrows:,} associated with SDSS magnitudes.")

        # Step 3: localise entries with empty magnitudes and use PanSTARRS1
        # transformations.

        # PS1 fluxes are in Janskys. We use stacked fluxes instead of mean
        # magnitudes since they are more complete on the faint end.
        ps1_g = 8.9 - 2.5 * peewee.fn.log(cdb.Panstarrs1.g_stk_psf_flux)
        ps1_r = 8.9 - 2.5 * peewee.fn.log(cdb.Panstarrs1.r_stk_psf_flux)
        ps1_i = 8.9 - 2.5 * peewee.fn.log(cdb.Panstarrs1.i_stk_psf_flux)

        # Use transformations to SDSS from Tonry et al. 2012 (section 3.2, table 6).
        x = ps1_g - ps1_r
        ps1_sdss_g = 0.013 + 0.145 * x + 0.019 * x * x + ps1_g
        ps1_sdss_r = -0.001 + 0.004 * x + 0.007 * x * x + ps1_r
        ps1_sdss_i = -0.005 + 0.011 * x + 0.010 * x * x + ps1_i

        # limit panstarrs photometry to r > 14
        max_flux_r = 10 ** ((14.0 - 8.9) / -2.5)

        with self.database.atomic():
            self.database.execute_sql("DROP TABLE IF EXISTS " + self.table_name + "_ps1")
            temp_table = peewee.Table(self.table_name + "_ps1")

            (
                Model.select(
                    Model.catalogid,
                    ps1_sdss_g.alias("ps1_sdss_g"),
                    ps1_sdss_r.alias("ps1_sdss_r"),
                    ps1_sdss_i.alias("ps1_sdss_i"),
                )
                .join(
                    cdb.CatalogToPanstarrs1,
                    on=(cdb.CatalogToPanstarrs1.catalogid == Model.catalogid),
                )
                .join(
                    cdb.Panstarrs1,
                    on=(cdb.CatalogToPanstarrs1.target_id == cdb.Panstarrs1.catid_objid),
                )
                .where(Model.g.is_null() | Model.r.is_null() | Model.i.is_null())
                .where(Model.selected >> True)
                .where(
                    cdb.CatalogToPanstarrs1.best >> True,
                    cdb.CatalogToPanstarrs1.version_id == self.get_version_id(),
                )
                .where(cdb.Panstarrs1.r_stk_psf_flux < max_flux_r)
                .where(
                    cdb.Panstarrs1.g_stk_psf_flux.is_null(False),
                    cdb.Panstarrs1.g_stk_psf_flux != "NaN",
                    (cdb.Panstarrs1.g_stk_psf_flux > 0),
                )
                .where(
                    cdb.Panstarrs1.r_stk_psf_flux.is_null(False),
                    cdb.Panstarrs1.r_stk_psf_flux != "NaN",
                    (cdb.Panstarrs1.r_stk_psf_flux > 0),
                )
                .where(
                    cdb.Panstarrs1.i_stk_psf_flux.is_null(False),
                    cdb.Panstarrs1.i_stk_psf_flux != "NaN",
                    (cdb.Panstarrs1.i_stk_psf_flux > 0),
                )
                .create_table(temp_table._path[0], temporary=True)
            )

            nrows = (
                Model.update(
                    {
                        Model.g: temp_table.c.ps1_sdss_g,
                        Model.r: temp_table.c.ps1_sdss_r,
                        Model.i: temp_table.c.ps1_sdss_i,
                        Model.optical_prov: peewee.Value("sdss_psfmag_ps1"),
                    }
                )
                .from_(temp_table)
                .where(Model.catalogid == temp_table.c.catalogid)
                .where(
                    temp_table.c.ps1_sdss_g.is_null(False)
                    & temp_table.c.ps1_sdss_r.is_null(False)
                    & temp_table.c.ps1_sdss_i.is_null(False)
                )
            ).execute()

        self.log.debug(f"{nrows:,} associated with PS1 magnitudes.")

        # Step 4: localise entries with empty magnitudes and use Gaia transformations
        # from Evans et al (2018).

        gaia_G = cdb.Gaia_DR3.phot_g_mean_mag
        gaia_BP = cdb.Gaia_DR3.phot_bp_mean_mag
        gaia_RP = cdb.Gaia_DR3.phot_rp_mean_mag

        x = gaia_BP - gaia_RP
        x2 = x * x
        x3 = x * x * x
        gaia_sdss_g = -1 * (0.13518 - 0.46245 * x - 0.25171 * x2 + 0.021349 * x3) + gaia_G
        gaia_sdss_r = -1 * (-0.12879 + 0.24662 * x - 0.027464 * x2 - 0.049465 * x3) + gaia_G
        gaia_sdss_i = -1 * (-0.29676 + 0.64728 * x - 0.10141 * x2) + gaia_G

        with self.database.atomic():
            self.database.execute_sql("DROP TABLE IF EXISTS " + self.table_name + "_gaia")
            temp_table = peewee.Table(self.table_name + "_gaia")

            (
                Model.select(
                    Model.catalogid,
                    gaia_sdss_g.alias("gaia_sdss_g"),
                    gaia_sdss_r.alias("gaia_sdss_r"),
                    gaia_sdss_i.alias("gaia_sdss_i"),
                )
                .join(
                    cdb.CatalogToGaia_DR3,
                    on=(cdb.CatalogToGaia_DR3.catalogid == Model.catalogid),
                )
                .join(cdb.Gaia_DR3)
                .where(Model.g.is_null() | Model.r.is_null() | Model.i.is_null())
                .where(Model.selected >> True)
                .where(
                    cdb.CatalogToGaia_DR3.best >> True,
                    cdb.CatalogToGaia_DR3.version_id == self.get_version_id(),
                )
                .where(cdb.Gaia_DR3.phot_g_mean_mag.is_null(False))
                .where(cdb.Gaia_DR3.phot_bp_mean_mag.is_null(False))
                .where(cdb.Gaia_DR3.phot_rp_mean_mag.is_null(False))
                .create_table(temp_table._path[0], temporary=True)
            )

            nrows = (
                Model.update(
                    {
                        Model.g: temp_table.c.gaia_sdss_g,
                        Model.r: temp_table.c.gaia_sdss_r,
                        Model.i: temp_table.c.gaia_sdss_i,
                        Model.optical_prov: peewee.Value("sdss_psfmag_gaia"),
                    }
                )
                .from_(temp_table)
                .where(Model.catalogid == temp_table.c.catalogid)
                .where(
                    temp_table.c.gaia_sdss_g.is_null(False)
                    & temp_table.c.gaia_sdss_r.is_null(False)
                    & temp_table.c.gaia_sdss_i.is_null(False)
                )
            ).execute()

        self.log.debug(f"{nrows:,} associated with Gaia DR3 mean magnitudes.")

        # Finally, check if there are any rows in which at least some of the
        # magnitudes are null.

        n_empty = (
            Model.select()
            .where(Model.g.is_null() | Model.r.is_null() | Model.i.is_null())
            .where(Model.selected >> True)
            .count()
        )

        if n_empty > 0:
            log.warning(f"Found {n_empty} entries with empty magnitudes.")

    def post_process(self, model, **kwargs):
        """Post-processes the temporary table.

        This method provides a framework for applying non-SQL operations on
        carton query. It receives the model for the temporary table and can
        perform any operation on it, including modifying the ``selected``
        column with a mask of targets to be used.

        This method can also be used to set the ``cadence`` column in the
        temporary table. This column will be used to set the target cadence if
        the carton `.cadence` attribute is not set.

        `.post_process` runs inside a database transaction so it's not
        necessary to create a new one, but savepoints can be added.

        Parameters
        ----------
        model : peewee:Model
            The model of the intermediate table.

        Returns
        -------
        mask : `tuple`
            The list of catalogids from the temporary table that should be
            selected as part of this carton. If `True` (the default), selects
            all the records.

        """

        return True

    def setup_transaction(self):
        """Setups the transaction locally modifying the datbase parameters.

        This method runs inside a transaction and can be overridden to set the
        parameters of the transaction manually. It applies to both `.run` and
        `.load`.

        """

        if "database_options" not in self.config:
            return

        for param, value in self.config["database_options"].items():
            self.database.execute_sql(f"SET LOCAL {param} = {value!r};")

    def drop_table(self):
        """Drops the intermediate table if it exists."""

        self.database.execute_sql(f"DROP TABLE IF EXISTS {self.path};")

    def write_table(self, filename=None, mode="results", write=True):
        """Writes the selection to a FITS file.

        Parameters
        ----------
        filename : str
            The file to which to write the table. Defaults to
            ``<name>_<plan>.fits``.
        mode : str
            Defines what data to write. If ``'results'``, writes the
            intermediate table (usually just the ``catalogid`` column). If
            ``'targetdb'``, writes all the relevant columns for the targets
            loaded to ``targetdb`` for this carton and plan (must be used after
            `.load` has been called).
        write : bool
            Whether to write the table to disk. If `False`, just returns the
            table object.

        Returns
        -------
        table : `~astropy.table.Table`
            A table object with the selected results.

        """

        if not filename:
            if mode == "results":
                filename = f"{self.name}_{self.plan}.fits.gz"
            else:
                filename = f"{self.name}_{self.plan}_targetdb.fits.gz"

        self.log.debug(f"Writing table to {filename}.")

        if not self.RModel:
            self.database.execute_sql("COMMIT;")
            self.RModel = self.get_model()

        if mode == "results":
            results_model = self.RModel
            assert results_model.table_exists(), "temporary table does not exist."

            write_query = results_model.select().order_by(results_model.catalogid)

            colnames = [field.name for field in write_query._returning]

        elif mode == "targetdb":
            mag_fields = [
                field
                for field in tdb.Magnitude._meta.fields.values()
                if field.name not in ["pk", "target_pk", "target"]
            ]

            write_query = (
                tdb.Target.select(
                    tdb.Target,
                    *mag_fields,
                    tdb.CartonToTarget.priority,
                    tdb.CartonToTarget.value,
                    tdb.CartonToTarget.instrument_pk,
                    tdb.CartonToTarget.inertial,
                    tdb.CartonToTarget.can_offset,
                    tdb.Cadence.label.alias("cadence"),
                )
                .join(tdb.CartonToTarget)
                .join(tdb.Magnitude)
                .join_from(tdb.CartonToTarget, tdb.Cadence, peewee.JOIN.LEFT_OUTER)
                .join_from(tdb.CartonToTarget, tdb.Carton)
                .join(tdb.Version)
                .where(
                    tdb.Carton.carton == self.name,
                    tdb.Version.plan == self.plan,
                    tdb.Version.target_selection >> True,
                )
                .order_by(tdb.Target.catalogid)
            )

            colnames = []
            for col in write_query._returning:
                if isinstance(col, peewee.ForeignKeyField):
                    colnames.append(col.column_name)
                elif isinstance(col, peewee.Alias):
                    colnames.append(col._alias)
                else:
                    colnames.append(col.name)

        else:
            raise ValueError('invalud mode. Available modes are "results" and "targetdb".')

        if not write_query.exists():
            raise TargetSelectionError("no records found.")

        results = write_query.tuples()
        results = (
            (col if col is not None else numpy.nan for col in row) for row in tuple(results)
        )

        warnings.filterwarnings("ignore", message=".*converting a masked element to nan.*")

        carton_table = table.Table(rows=results, names=colnames, masked=True)

        if write:
            carton_table.write(filename, overwrite=True)

        return carton_table

    def load(self, mode="fail", overwrite=False):
        """Loads the output of the intermediate table into targetdb.

        Parameters
        ----------
        mode : str
            The mode to use when loading the targets. If ``'fail'``, raises an
            error if the carton already exist. If ``'overwrite'``, overwrites
            the targets. If ``'append'``, appends the targets.
        overwrite : bool
            Equivalent to setting ``mode='overwrite'``. This option is deprecated and
            will raise a warning.

        """

        if overwrite:
            mode = "overwrite"
            log.warning(
                "The `overwrite` option is deprecated and will be removed in a future version. "
                'Use `mode="overwrite"` instead.'
            )

        if self.check_targets():
            if mode == "overwrite":
                log.warning(
                    f"Carton {self.name!r} with plan {self.plan!r} "
                    f"already has targets loaded. "
                    "Dropping carton-to-target entries."
                )
                self.drop_carton()
            elif mode == "append":
                pass
            elif mode == "fail":
                raise TargetSelectionError(
                    f"Found existing targets for "
                    f"carton {self.name!r} with plan "
                    f"{self.plan!r}."
                )
            else:
                raise ValueError(f'Invalid mode {mode!r}. Use "fail", "overwrite", or "append".')

        if self.RModel is None:
            self.database.execute_sql("COMMIT;")
            RModel = self.get_model()
        else:
            RModel = self.RModel

        if not RModel.table_exists():
            raise TargetSelectionError(
                f"No temporary table found {self.full}. Did you call run()?"
            )

        has_targets = RModel.select().where(RModel.selected >> True).exists()

        if not has_targets:
            log.warning("No targets found in intermediate table.")

        with self.database.atomic():
            self.setup_transaction()
            self._create_carton_metadata()
            self._load_targets(RModel)
            self._load_carton_to_target(RModel)
            if self.load_magnitudes:
                self._load_magnitudes(RModel)
            else:
                log.warning("Skipping magnitude load.")

            self.log.debug("Committing records and checking constraints.")

    def check_targets(self):
        """Check if data has been loaded for this carton and targeting plan."""

        has_targets = (
            tdb.CartonToTarget.select()
            .join(tdb.Carton)
            .join(tdb.Version)
            .where(
                tdb.Carton.carton == self.name,
                tdb.Version.plan == self.plan,
                tdb.Version.target_selection >> True,
            )
            .exists()
        )

        return has_targets

    def _create_carton_metadata(self):
        mapper_pk = None
        category_pk = None

        # Create targeting plan in tdb.
        version, created = tdb.Version.get_or_create(
            plan=self.plan,
            tag=self.tag,
            target_selection=True,
            robostrategy=False,
        )
        version_pk = version.pk

        if created:
            self.log.info(
                f"Created record in targetdb.version for {self.plan!r} with tag {self.tag!r}."
            )

        if (
            tdb.Carton.select()
            .where(tdb.Carton.carton == self.name, tdb.Carton.version_pk == version_pk)
            .exists()
        ):
            return

        # Create carton and associated values.
        if self.mapper:
            mapper, created_pk = tdb.Mapper.get_or_create(label=self.mapper)
            mapper_pk = mapper.pk
            if created:
                self.log.debug(f"Created mapper {self.mapper!r}")

        if self.category:
            category, created = tdb.Category.get_or_create(label=self.category)
            category_pk = category.pk
            if created:
                self.log.debug(f"Created category {self.category!r}")

        tdb.Carton.create(
            carton=self.name,
            category_pk=category_pk,
            program=self.program,
            mapper_pk=mapper_pk,
            version_pk=version_pk,
            run_on=datetime.datetime.now().isoformat().split("T")[0],
        ).save()

        self.log.debug(f"Created carton {self.name!r}")

    def _load_targets(self, RModel):
        """Load data from the intermediate table tp targetdb.target."""

        self.log.debug("loading data into targetdb.target.")

        (
            tdb.Target.insert_from(
                cdb.Catalog.select(
                    cdb.Catalog.catalogid,
                    cdb.Catalog.ra,
                    cdb.Catalog.dec,
                    cdb.Catalog.pmra,
                    cdb.Catalog.pmdec,
                    cdb.Catalog.parallax,
                    peewee.Value(EPOCH),
                )
                .join(RModel, on=(cdb.Catalog.catalogid == RModel.catalogid))
                .where(RModel.selected >> True)
                .where(
                    ~peewee.fn.EXISTS(
                        tdb.Target.select(peewee.SQL("1")).where(
                            tdb.Target.catalogid == RModel.catalogid
                        )
                    )
                ),
                [
                    tdb.Target.catalogid,
                    tdb.Target.ra,
                    tdb.Target.dec,
                    tdb.Target.pmra,
                    tdb.Target.pmdec,
                    tdb.Target.parallax,
                    tdb.Target.epoch,
                ],
            )
            .returning()
            .execute()
        )

        self.log.info("Inserted new rows into targetdb.target.")

        return

    def _load_magnitudes(self, RModel):
        """Load magnitudes into targetdb.magnitude."""

        self.log.debug("Loading data into targetdb.magnitude.")

        Magnitude = tdb.Magnitude

        magnitude_paths = self.config["magnitudes"]
        fields = [Magnitude.carton_to_target_pk]

        select_from = (
            RModel.select(tdb.CartonToTarget.pk)
            .join(tdb.Target, on=(RModel.catalogid == tdb.Target.catalogid))
            .join(tdb.CartonToTarget)
            .join(tdb.Carton)
            .join(tdb.Version)
            .where(RModel.selected >> True)
            .where(
                tdb.Carton.carton == self.name,
                tdb.Version.plan == self.plan,
                tdb.Version.tag == self.tag,
                tdb.Version.target_selection >> True,
            )
        )

        for mag, mpath in magnitude_paths.items():
            fields.append(getattr(Magnitude, mag))
            if hasattr(RModel, mag):
                select_from = select_from.select_extend(getattr(RModel, mag))
                continue

            select_from = select_from.switch(tdb.Target)

            # For each node in the join list we check if the node model has
            # already been join and if so, switch the pointer to that model.
            # Otherwise do a LEFT OUTER join because we want all the rows in
            # the temporary table even if they don't have associated
            # magnitudes. We make sure we only use the "best" match from the
            # cross-match. No need to apply filters on cross-match version_id
            # because catalogid is unique across versions.

            for node in mpath:
                column = None
                if node == mpath[-1]:
                    node, column = node.split(".")
                node_model = self.database.models.get("catalogdb." + node)
                joins = [model[0] for join in select_from._joins.values() for model in join]
                if node_model in joins:
                    select_from = select_from.switch(node_model)
                else:
                    if node.startswith("catalog_to_"):
                        select_from = select_from.join(
                            node_model,
                            peewee.JOIN.LEFT_OUTER,
                            on=(tdb.Target.catalogid == node_model.catalogid),
                        )
                        select_from = select_from.where(
                            (
                                (node_model.best >> True)
                                & (node_model.version_id == self.get_version_id())
                            )
                            | (node_model.catalogid >> None)
                        )
                    else:
                        select_from = select_from.join(node_model, peewee.JOIN.LEFT_OUTER)
                if column:
                    select_from = select_from.select_extend(getattr(node_model, column))

        # Add gri from the temporary table.
        for mag in ["g", "r", "i", "z"]:
            select_from = select_from.select_extend(RModel._meta.columns[mag])
            fields.append(Magnitude._meta.columns[mag])

        if "optical_prov" not in RModel._meta.columns:
            raise TargetSelectionError("optical_prov column not found in temporary table.")

        select_from = select_from.select_extend(RModel._meta.columns["optical_prov"])
        fields.append(Magnitude.optical_prov)

        Magnitude.insert_from(select_from, fields).returning().execute()

        self.log.info("Inserted new rows into targetdb.magnitude.")

    def _load_carton_to_target(self, RModel):
        """Populate targetdb.carton_to_target."""

        self.log.debug("Loading data into targetdb.carton_to_target.")

        version_pk = tdb.Version.get(
            plan=self.plan,
            tag=self.tag,
            target_selection=True,
        )
        carton_pk = tdb.Carton.get(carton=self.name, version_pk=version_pk).pk

        Target = tdb.Target
        CartonToTarget = tdb.CartonToTarget

        select_from = (
            RModel.select(Target.pk, carton_pk)
            .join(Target, on=(Target.catalogid == RModel.catalogid))
            .where(RModel.selected >> True)
            .where(
                ~peewee.fn.EXISTS(
                    CartonToTarget.select(peewee.SQL("1"))
                    .join(tdb.Carton)
                    .where(
                        CartonToTarget.target_pk == Target.pk,
                        CartonToTarget.carton_pk == carton_pk,
                        tdb.Carton.version_pk == version_pk,
                    )
                )
            )
        )

        if self.cadence is not None:
            # Check that not both the carton cadence and the cadence column
            # are not null.
            if "cadence" in RModel._meta.fields:
                if RModel.select().where(~(RModel.cadence >> None)).exists():
                    raise TargetSelectionError(
                        "both carton cadence and target "
                        "cadence defined. This is not "
                        "allowed."
                    )

            cadence_pk = tdb.Cadence.get(label=self.cadence)
            select_from = select_from.select_extend(cadence_pk)

            if not self.value:
                # improve robustness of cadence name pattern matching slightly:
                try:
                    chunks = [s for s in self.cadence.split("_") if re.match(r"[0-9]+x[0-9]+", s)]
                    cadence_payload = chunks[0].split("x")
                except BaseException:
                    raise ("Uninterpretable cadence name: ", self.cadence)

                self.value = float(
                    numpy.multiply(
                        # *map(int, self.cadence.split('_')[-1].split('x'))
                        *map(int, cadence_payload)
                    )
                )

        else:
            # If all cadences are null we'll set that as a value and save us
            # a costly join.
            if not RModel.select().where(~(RModel.cadence >> None)).exists():
                select_from = select_from.select_extend(peewee.SQL("null"))
            else:
                select_from = (
                    select_from.select_extend(tdb.Cadence.pk)
                    .switch(RModel)
                    .join(
                        tdb.Cadence,
                        "LEFT OUTER JOIN",
                        on=(tdb.Cadence.label == RModel.cadence),
                    )
                )

        if self.priority is None:
            select_from = select_from.select_extend(RModel.priority)
        else:
            select_from = select_from.select_extend(self.priority)

        if self.can_offset is None:
            select_from = select_from.select_extend(RModel.can_offset)
        else:
            select_from = select_from.select_extend(self.can_offset)

        if self.value is not None:
            select_from = select_from.select_extend(self.value)
        else:
            # We will use the cadence to determine the value. First, if there is
            # not a user-defined value column, create it.
            if "value" not in RModel._meta.columns:
                self.database.execute_sql(f"ALTER TABLE {self.path} ADD COLUMN value REAL;")

                # We need to add the field like this and not call get_model() because
                # at this point the temporary table is locked and reflection won't work.
                RModel._meta.add_field("value", peewee.FloatField())

            # Get the value as the n_epochs * n_exposures_per_epoch. Probably this can
            # be done directly in SQL but it's just easier in Python. Note that because
            # we set value above in the case when cadence is a single value, if we
            # are here that means there is a cadence column.

            data = numpy.array(
                RModel.select(RModel.catalogid, RModel.cadence)
                .where(RModel.cadence.is_null(False))
                .tuples()
            )

            def split_cadence(cadence):
                chunks = [s for s in cadence.split("_") if re.match(r"[0-9]+x[0-9]+", s)]
                return chunks[0].split("x")

            if data.size > 0:
                values = tuple(
                    int(
                        numpy.multiply(
                            # *map(int, cadence.split('_')[-1].split('x'))))
                            # improve robustness of cadence name pattern matching slightly:
                            *map(int, split_cadence(cadence))
                        )
                    )
                    for cadence in data[:, 1]
                )

                catalogid_values = zip(map(int, data[:, 0]), values)

                vl = peewee.ValuesList(
                    catalogid_values,
                    columns=("catalogid", "value"),
                    alias="vl",
                )

                (
                    RModel.update(value=vl.c.value)
                    .from_(vl)
                    .where(RModel.catalogid == vl.c.catalogid)
                    .where(RModel.value.is_null())
                ).execute()

            select_from = select_from.select_extend(RModel.value)

        if "instrument" in RModel._meta.columns:
            select_from = (
                select_from.select_extend(tdb.Instrument.pk)
                .switch(RModel)
                .join(
                    tdb.Instrument,
                    "LEFT OUTER JOIN",
                    on=(tdb.Instrument.label == RModel.instrument),
                )
            )
        elif self.instrument is not None:
            select_from = select_from.select_extend(tdb.Instrument.get(label=self.instrument).pk)
        else:
            raise RuntimeError(f"Instrument not defined for carton {self.name}")

        for colname in ["delta_ra", "delta_dec", "inertial"]:
            if colname in RModel._meta.columns:
                select_from = select_from.select_extend(RModel._meta.columns[colname])
            else:
                if colname == "inertial":
                    select_from = select_from.select_extend(peewee.Value(False))
                else:
                    select_from = select_from.select_extend(peewee.Value(0.0))

        if "lambda_eff" in RModel._meta.columns:
            select_from = select_from.select_extend(RModel._meta.columns["lambda_eff"])
        else:
            if self.instrument is not None:
                instrument = self.instrument
            else:
                instrument = RModel.instrument
            select_from = select_from.select_extend(
                tdb.Instrument.select(tdb.Instrument.default_lambda_eff).where(
                    tdb.Instrument.label == instrument
                )
            )

        # Now do the insert
        (
            CartonToTarget.insert_from(
                select_from,
                [
                    CartonToTarget.target_pk,
                    CartonToTarget.carton_pk,
                    CartonToTarget.cadence_pk,
                    CartonToTarget.priority,
                    CartonToTarget.can_offset,
                    CartonToTarget.value,
                    CartonToTarget.instrument_pk,
                    CartonToTarget.delta_ra,
                    CartonToTarget.delta_dec,
                    CartonToTarget.inertial,
                    CartonToTarget.lambda_eff,
                ],
            )
            .returning()
            .execute()
        )

        self.log.info("Inserted rows into targetdb.carton_to_target.")

    def drop_carton(self):
        """Drops the entry in ``targetdb.carton``."""

        version = tdb.Version.select().where(
            tdb.Version.plan == self.plan,
            tdb.Version.tag == self.tag,
            tdb.Version.target_selection >> True,
        )

        if version.count() == 0:
            return

        tdb.Carton.delete().where(
            tdb.Carton.carton == self.name, tdb.Carton.version == version
        ).execute()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2020-06-16
# @Filename: mwm_dust.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

import numpy
import pandas
import peewee

from sdssdb.peewee.sdss5db import targetdb
from sdssdb.peewee.sdss5db.catalogdb import (
    GLIMPSE,
    AllWise,
    Catalog,
    CatalogToAllWise,
    CatalogToGaia_DR3,
    CatalogToGLIMPSE,
    CatalogToTwoMassPSC,
    Gaia_DR3,
    TwoMassPSC,
)

from target_selection.cartons import BaseCarton


def lbp2xyz(ll, bb, pp):
    dist = 1.0 / pp  # kpc
    z = dist * numpy.sin(numpy.radians(bb))
    x = dist * numpy.cos(numpy.radians(bb)) * numpy.cos(numpy.radians(ll))
    y = dist * numpy.cos(numpy.radians(bb)) * numpy.sin(numpy.radians(ll))

    # z is definitely right; x and y depend on conventions.  As long
    # as this routine is used for x and y throughout this code, this should
    # not introduce any issues.

    return x, y, z


# subselect() is used in post_process() to randomly select a subset
# from the results of the  query in build_query().
# Hence, the query in build_query() uses order_by().
# The order_by() in the query ensures that the random selection in
# post_process() gives the same result every time we run this carton.


def subselect(data, othersel, downsampledby=1):
    """Turn x, y, z into pixel; bin up to see where we are missing."""

    ll = data["gallong"]
    bb = data["gallat"]
    xyz = lbp2xyz(ll, bb, data["parallax"])

    dustsel = numpy.ones(len(data), numpy.bool)

    ranges = [[-5, 5], [-5, 5], [-0.2, 0.2]]
    resolution = 0.1

    countshape = [numpy.round((r[1] - r[0]) / resolution).astype("i4") + 2 for r in ranges]

    coords = [
        numpy.clip(numpy.floor((c - r[0]) / resolution).astype("i4"), -1, s - 2) + 1
        for (c, r, s) in zip(xyz, ranges, countshape)
    ]
    centers = [
        r[0] + resolution * (numpy.arange((r[1] - r[0]) / resolution) + 0.5) for r in ranges
    ]

    flatcoord = numpy.ravel_multi_index(coords, countshape)

    countsd = numpy.bincount(
        flatcoord[dustsel & ~othersel],
        minlength=numpy.product(countshape),
    )
    countso = numpy.bincount(
        flatcoord[othersel & dustsel],
        minlength=numpy.product(countshape),
    )
    countsd = countsd.reshape(countshape)
    countso = countso.reshape(countshape)

    pnumdensd = map_coordinates_wrap(
        (countsd[1:-1, 1:-1, 1:-1], centers),
        [c[dustsel & ~othersel] for c in xyz],
        order=0,
    )
    pnumdenso = map_coordinates_wrap(
        (countso[1:-1, 1:-1, 1:-1], centers),
        [c[dustsel & ~othersel] for c in xyz],
        order=0,
    )

    ntopick = 10.0 / downsampledby
    pnumdensleft = ntopick - pnumdenso

    numpy.random.seed(0)
    randnum = numpy.random.rand(numpy.sum(dustsel & ~othersel))

    subsel = randnum < pnumdensleft / 1.0 / (pnumdensd + (pnumdensd == 0))

    dustsel[othersel] = 0
    dustsel[dustsel & ~othersel] = subsel

    return dustsel


def map_coordinates_wrap(grid, coord, **kw):
    gridpts = grid[0]
    gridcoord = grid[1]
    outputcoordnorm = [
        numpy.interp(c, gc, numpy.arange(len(gc))) for c, gc in zip(coord, gridcoord)
    ]
    from scipy.ndimage import interpolation

    return interpolation.map_coordinates(
        gridpts,
        outputcoordnorm,
        cval=0.0,
        mode="constant",
        **kw,
    )


# These were initial selection criteria, not used anymore. We use the
# real GG carton instead.


def jkselect(data):
    jk0 = data["j_ks_0"]
    return (jk0 > 0.7) & (data["hmag"] < 11)


def ghselect(data):
    gaiag = data["gaiamag"]
    return (data["hmag"] < 11) & numpy.where(
        numpy.isfinite(gaiag),
        gaiag - data["hmag"] > 3.5,
        True,
    )


# The below MWM_Dust_Core_apogee_Carton carton is obsolete.
# Do not run this carton for v1.0.
# However, do not delete this carton since
# this carton is the base carton for MWM_Dust_Core_Dist_apogee_Carton.
# This carton requires that the MWM_Galatic_Core_apogee_Carton
# has been run before it in the same target_selection_plan
# in target_selection.yml or in an earlier target_selection_plan.
# See target_selection.yml for example for mwm_dust_core_apogee.
class MWM_Dust_Core_apogee_Carton(BaseCarton):
    """MWM Dust Core apogee Carton.

    Definition:

        A sample of bright, nearby, midplane giants that, when combined
        with the more distant and/or heavily reddened sample in other cartons,
        produces 100 stars per (100pc)^3 volume in the near disk.

        - Use Gaia parallax and (l,b) to get distance
        - Use H and 4.5 micron to get A_Ks and E_JKs, a la APOGEE
          (using RJCE dereddening prescription from Majewski et al. 2011).
        - Use J, K, and E_JKs to get (J-Ks)_0
        - Use distance, K, and A_Ks to get absolute mag M_K
        - 0 < Gaia parallax error/parallax < 0.2
        - M_K < 2.6
        - H < 11.2
        - distance < 5 kpc
        - |z| < 0.2 kpc
        - (J-Ks)_0 > 0.5
        - Can be spatially subselected to complement GG sampling in order to
          obtain 100 stars per (100pc)^3 volume.
        - Galactic Genesis quality flags:
            (ph_qual[1] = 'A' OR ph_qual[1] = 'B')
                AND gal_contam=0
                AND cc_flg[1]=0
                AND (rd_flg[1] > 0 AND rd_flg[1] <= 3)

    Non-SQL implementation:

        https://faun.rc.fas.harvard.edu/eschlafly/sdss5/dustsel.py

    """

    name = "mwm_dust_core_apogee"
    mapper = "MWM"
    category = "science"
    program = "mwm_dust"
    instrument = "APOGEE"
    cadence = "bright_1x1"
    priority = 2720
    can_offset = True

    mwm_galactic_carton = None
    mwm_galactic_plan = None

    def build_query(self, version_id, query_region=None):
        # Get the plan used to run mwm_galactic_core_apogee.
        if self.name == "mwm_dust_core_apogee":
            self.mwm_galactic_carton = "mwm_galactic_core_apogee"
        elif self.name == "mwm_dust_core_dist_apogee":
            self.mwm_galactic_carton = "mwm_galactic_core_dist_apogee"
        else:
            raise ValueError(f"Invalid carton {self.name}.")

        if self.parameters:
            self.mwm_galactic_plan = self.parameters.get(
                self.mwm_galactic_carton + "_plan", self.plan
            )
        else:
            self.mwm_galactic_plan = self.plan

        gg_exists = (
            targetdb.Carton.select()
            .join(targetdb.Version)
            .where(
                targetdb.Carton.carton == self.mwm_galactic_carton,
                targetdb.Version.plan == self.mwm_galactic_plan,
                targetdb.Version.target_selection >> True,
            )
            .exists()
        )
        if not gg_exists:
            raise RuntimeError(self.mwm_galactic_carton + " has not been loaded yet.")

        fn = peewee.fn

        # GG quality flags
        ph_qual = TwoMassPSC.ph_qual
        cc_flg = TwoMassPSC.cc_flg
        rd_flg = TwoMassPSC.rd_flg
        rd_flag_1 = peewee.fn.substr(rd_flg, 2, 1).cast("integer")
        gal_contam = TwoMassPSC.gal_contam

        gallong = Gaia_DR3.l
        gallat = Gaia_DR3.b

        ipar = 1.0 / Gaia_DR3.parallax  # kpc
        zz = ipar * fn.sin(fn.radians(gallat))
        xx = ipar * fn.cos(fn.radians(gallat)) * fn.cos(fn.radians(gallong))
        yy = ipar * fn.cos(fn.radians(gallat)) * fn.sin(fn.radians(gallong))

        dist = fn.sqrt(fn.pow(xx, 2) + fn.pow(yy, 2) + fn.pow(zz, 2))

        aks_glimpse = 0.918 * (GLIMPSE.mag_h - GLIMPSE.mag4_5 - 0.08)
        aks_allwise = 0.918 * (AllWise.h_m_2mass - AllWise.w2mpro - 0.08)
        aks = fn.coalesce(aks_glimpse, aks_allwise)

        Ej_ks = 1.5 * aks

        j_ks_0_glimpse = GLIMPSE.mag_j - GLIMPSE.mag_ks - Ej_ks
        j_ks_0_allwise = AllWise.j_m_2mass - AllWise.k_m_2mass - Ej_ks
        j_ks_0 = fn.coalesce(j_ks_0_glimpse, j_ks_0_allwise)

        plxfracunc = 1 / Gaia_DR3.parallax_over_error
        dm = 5 * fn.log(1000.0 / Gaia_DR3.parallax / 10)
        absmag = TwoMassPSC.k_m - aks - dm

        query = (
            Gaia_DR3.select(
                CatalogToGaia_DR3.catalogid,
                peewee.Value(False).alias("selected"),  # Set selected to False
                Gaia_DR3.source_id.alias("gaia_souce_id"),
                Gaia_DR3.l.alias("gallong"),
                Gaia_DR3.b.alias("gallat"),
                Gaia_DR3.parallax,
                Gaia_DR3.phot_g_mean_mag,
                TwoMassPSC.h_m,
                TwoMassPSC.k_m,
                aks.alias("a_ks"),
                j_ks_0.alias("j_ks_0"),
            )
            .join(CatalogToGaia_DR3)
            .join(
                CatalogToAllWise,
                peewee.JOIN.LEFT_OUTER,
                on=(CatalogToAllWise.catalogid == CatalogToGaia_DR3.catalogid),
            )
            .join(AllWise, peewee.JOIN.LEFT_OUTER)
            .join_from(
                CatalogToGaia_DR3,
                CatalogToTwoMassPSC,
                peewee.JOIN.LEFT_OUTER,
                on=(CatalogToTwoMassPSC.catalogid == CatalogToGaia_DR3.catalogid),
            )
            .join(TwoMassPSC, peewee.JOIN.LEFT_OUTER)
            .join_from(
                CatalogToAllWise,
                CatalogToGLIMPSE,
                peewee.JOIN.LEFT_OUTER,
                on=(CatalogToGLIMPSE.catalogid == CatalogToAllWise.catalogid),
            )
            .join(GLIMPSE, peewee.JOIN.LEFT_OUTER)
            .where(
                ((CatalogToAllWise.version_id == version_id) & (CatalogToAllWise.best >> True))
                | (CatalogToAllWise.catalogid >> None)
            )
            .where(
                CatalogToGaia_DR3.version_id == version_id,
                CatalogToGaia_DR3.best >> True,
            )
            .where(
                CatalogToTwoMassPSC.version_id == version_id,
                CatalogToTwoMassPSC.best >> True,
            )
            .where(
                ((CatalogToGLIMPSE.version_id == version_id) & (CatalogToGLIMPSE.best >> True))
                | (CatalogToGLIMPSE.catalogid >> None)
            )
            .where(
                TwoMassPSC.h_m < 11.2,
                fn.abs(zz) < 0.2,
                j_ks_0.is_null(False),
                j_ks_0 > 0.5,
                dist < 5,
                plxfracunc < 0.2,
                plxfracunc > 0,
                absmag < 2.6,
            )
            .where(
                ph_qual.regexp(".(A|B)."),
                gal_contam == 0,
                peewee.fn.substr(cc_flg, 2, 1) == "0",
                rd_flag_1 > 0,
                rd_flag_1 <= 3,
            )
            .order_by(CatalogToGaia_DR3.catalogid)
        )

        # The above order_by() ensures that the random selection in
        # post_process() gives the same result every time we run this carton.

        if query_region:
            query = query.join_from(CatalogToAllWise, Catalog).where(
                peewee.fn.q3c_radial_query(
                    Catalog.ra,
                    Catalog.dec,
                    query_region[0],
                    query_region[1],
                    query_region[2],
                )
            )

        return query

    def post_process(self, model):
        """Select samples based on GG using Eddie Schlafly code.

        Goal subsample and to select only targets that are not already part
        of the Galactic Genesis sample. We use the real GG sample from
        targetdb.

        """

        data = pandas.read_sql(
            f"SELECT * FROM {model._meta.schema}.{model._meta.table_name};",
            self.database,
        )

        mwm_galactic = (
            targetdb.Target.select(targetdb.Target.catalogid)
            .join(targetdb.CartonToTarget)
            .join(targetdb.Carton)
            .join(targetdb.Version)
            .where(
                targetdb.Carton.carton == self.mwm_galactic_carton,
                targetdb.Version.plan == self.mwm_galactic_plan,
                targetdb.Version.target_selection >> True,
            )
        )

        gg_catids = tuple(zip(*mwm_galactic.tuples()))[0]
        gg_mask = numpy.in1d(data.catalogid, gg_catids)

        # subselect() is used below to randomly select a subset
        # from the results of the  query in build_query().
        # Hence, the query in build_query() uses order_by().
        # The order_by() in the query ensures that the random selection in
        # subselect() gives the same result every time we run this carton.
        #
        # Subsample based on GG and update DB.
        dust_gg_subsel = subselect(data, gg_mask)
        dust_gg_cid = peewee.ValuesList(
            zip(data.catalogid[dust_gg_subsel]),
            columns=("catalogid",),
            alias="vl",
        )

        (
            model.update({model.selected: True})
            .from_(dust_gg_cid)
            .where(model.catalogid == dust_gg_cid.c.catalogid)
            .execute()
        )

        return


# The below MWM_Dust_Core_Dist_apogee_Carton carton requires
# that the MWM_Galatic_Core_Dist_apogee_Carton
# has been run before this carton in the same target_selection_plan
# in target_selection.yml or in an earlier target_selection_plan.
# See target_selection.yml for example for mwm_dust_core_dist_apogee.
class MWM_Dust_Core_Dist_apogee_Carton(MWM_Dust_Core_apogee_Carton):
    """MWM Dust Core Dist apogee Carton.

    Same definition as the MWM_Dust_Core_apogee_Carton carton but
     subsampling using the mwm_galactic_core_dist_apogee.

    """

    name = "mwm_dust_core_dist_apogee"



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: Pramod Gupta (psgupta@uw.edu)
# @Date: 2020-06-10
# @Filename: mwm_rv.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

import math

import peewee

from sdssdb.peewee.sdss5db.catalogdb import (
    Catalog,
    CatalogToGaia_DR3,
    CatalogToTIC_v8,
    CatalogToTwoMassPSC,
    Gaia_DR3,
    SDSS_DR17_APOGEE_Allstarmerge,
    TIC_v8,
    TwoMassPSC,
)

from target_selection.cartons import BaseCarton


# See catalog.py for the name of peewee model names corresponding
# to postgres table names:
# https://github.com/sdss/sdssdb/blob/master/python/sdssdb/peewee/sdss5db/catalogdb.py

# 2.2.1. Long Baseline (Legacy Targets)

# Shorthand name: mwm_rv_long (Not an actual target class.
# Defined as a parent sample to select from for 2.2.1.x sections below)

# Simplified Description of selection criteria:
# APOGEE-1&-2 "main survey" targets with at least 3 previous epochs
# brighter than H of 12.8 that have a Gaia parallax.

# Wiki page: Binaries and Substellar Companions

# Additional source catalogs needed: Just need sdss_dr17_apogee_allstarmerge

# Additional cross-matching needed: (None)

# Return columns: apogee_id, nvisits, ra, dec, pmra, pmdec, h, baseline, fields

# cadence options for these targets (list all options,
# even though no single target will receive more than one):
# (See entries for this in 2.2.1.x below)

# Pseudo SQL (optional): SELECT apogee_id, nvisits, ra, dec,
# pmra, pmdec, h, baseline, fields FROM sdss_dr17_apogee_allstarmerge
# WHERE h<12.8 AND nvisits>=3 AND dist_src == 'GaiaDR3' AND
# [targflags contains 'APOGEE_SHORT' OR 'APOGEE_INTERMEDIATE' OR
# 'APOGEE_LONG' OR '*APOGEE2_*BIN_’] AS mwm_rv_long

# Implementation:

# Non-SQL implementation:

# lead contact: Nicholas Troup

# We associate each APOGEE target with a 2MASS source since the
# vast majority of APOGEE targets were selected from the 2MASS PSC.
# APOGEE_ID is essentially the same as
# the “designation” column of the 2MASS PSC;
# For sdss_dr17_apogee_allstarmerge we
# have to strip off the “2M” in the APOGEE_ID.
# However, some apogee_id values do not have 2M in the front.
# (old: for sdss_apogeeallstarmerge_r13 we also had to strip off the
#  “2M” in the APOGEE_ID.)
# For example:
# sdss5db=# select designation from  catalogdb.twomass_psc limit 2;
#    designation
# ------------------
#  12034281-5740002
#  12032366-5738380
# (2 rows)
#
# sdss5db=# select apogee_id from catalogdb.SDSS_DR17_APOGEE_Allstarmerge
# where apogee_id like '2M%';
#      apogee_id
# ---------------------
# 2M00000002+7417074
# 2M00000019-1924498
# etc.
# sdss5db=# select apogee_id from catalogdb.SDSS_DR17_APOGEE_Allstarmerge
# where apogee_id not like '2M%';
#    apogee_id
# --------------------
# 19140272-1554055
# 19155129-1617591
#
# #### start comment for old sdss_apogeeallstarmerge_r13 ###################
#
# sdss5db=# select replace(apogee_id,'2M', '') from
#  catalogdb.sdss_apogeeallstarmerge_r13 limit 2;
#       replace
# ------------------
#  14044120-1550575
#  14033676-1554164
# (2 rows)
#
#
# sdss5db=# select count(1) from
#  catalogdb.sdss_apogeeallstarmerge_r13 where dist_src='gaia';
#  count
# -------
#      0
# (1 row)
#
# Hence use trim(dist_src) like below:
#
# sdss5db=# select count(1) from
#  catalogdb.sdss_apogeeallstarmerge_r13 where trim(dist_src)='gaia';
#  count
# --------
#  487508
# (1 row)
#
# ######### end comment for old sdss_apogeeallstarmerge_r13 ##########

# Below we use GaiaEDR3 and not GaiaDR3 since the dist_src column
# does not have GaiaDR3

mwm_rv_long_condition = (
    SDSS_DR17_APOGEE_Allstarmerge.h < 11.5,  # old 12.8,
    SDSS_DR17_APOGEE_Allstarmerge.nvisits >= 6,  # old 3,
    peewee.fn.trim(SDSS_DR17_APOGEE_Allstarmerge.dist_src) == "GaiaEDR3",
    (SDSS_DR17_APOGEE_Allstarmerge.targflags % "%APOGEE_SHORT%")
    | (SDSS_DR17_APOGEE_Allstarmerge.targflags % "%APOGEE_INTERMEDIATE%")
    | (SDSS_DR17_APOGEE_Allstarmerge.targflags % "%APOGEE_LONG%")
    | (SDSS_DR17_APOGEE_Allstarmerge.targflags % "%APOGEE2_%BIN_%"),
)


# old class name MWM_RV_Long_FPS_Carton(BaseCarton):
class MWM_bin_rv_long_apogee_Carton(BaseCarton):
    """3.2.1.3. Long Baseline Targets for FPS

    Shorthand name: mwm_bin_rv_long_apogee (old name mwm_rv_long_fps)

    Simplified Description of selection criteria:
     Select from long-baseline targets (above) with H brighter than 11.5

    Wiki page: Binaries and Substellar Companions (Page Under Construction)

    Additional source catalogs needed: Select from mwm_rv_long (2.2.1)

    Additional cross-matching needed: (None)

    Return columns: apogee_id, nvisits, ra, dec,
     pmra, pmdec, h, baseline, fields (Same as 2.2.1 above)

    cadence options for these targets
     (list all options, even though no single target will receive more than one):
      If H>10.8 then use mwm_rv_<nn>x2, otherwise use mwm_rv_<nn>x1,
       where <nn> = 3*ceiling((18-nvisits)/3)
      if <nn> is less than 6 then
         set <nn> = 6

    Pseudo SQL (optional): SELECT apogee_id, nvisits, ra, dec,
     pmra, pmdec, h, baseline, fields FROM mwm_rv_long WHERE h<11.5

    Implementation:

    Non-SQL implementation:

    lead contact: Nicholas Troup

    """

    # cadence must be None here so that
    # it can be set in post_process().
    # If cadence is not None here then
    # it cannot be set in post_process().
    name = "mwm_bin_rv_long_apogee"  # old name = 'mwm_rv_long_fps'
    category = "science"
    instrument = "APOGEE"
    cadence = None  # cadence is set in post_process()
    program = "mwm_bin"
    mapper = "MWM"
    priority = 1810
    can_offset = True

    # peewee Model name ---> postgres table name
    # SDSS_DR17_APOGEE_Allstarmerge(CatalogdbModel)--->'sdss_dr17_apogee_allstarmerge'
    # There is no gaia_dr3 in the below query so for v1.0 we do not have to do
    # a major modification of the v0.5 query.
    # In the below query, we use replace() instead of ltrim() since
    # ltrim('2M20', '2M') will also trim the second 2.

    def build_query(self, version_id, query_region=None):
        query = (
            Catalog.select(
                CatalogToTIC_v8.catalogid,
                SDSS_DR17_APOGEE_Allstarmerge.apogee_id,
                SDSS_DR17_APOGEE_Allstarmerge.nvisits,
                SDSS_DR17_APOGEE_Allstarmerge.ra.alias("allstarmerge_ra"),
                SDSS_DR17_APOGEE_Allstarmerge.dec.alias("allstarmerge_dec"),
                SDSS_DR17_APOGEE_Allstarmerge.pmra.alias("allstarmerge_pmra"),
                SDSS_DR17_APOGEE_Allstarmerge.pmdec.alias("allstarmerge_pmdec"),
                SDSS_DR17_APOGEE_Allstarmerge.h,
                SDSS_DR17_APOGEE_Allstarmerge.baseline,
                SDSS_DR17_APOGEE_Allstarmerge.fields,
                SDSS_DR17_APOGEE_Allstarmerge.teff_avg,  # old teff
                SDSS_DR17_APOGEE_Allstarmerge.logg_avg,  # old logg
                SDSS_DR17_APOGEE_Allstarmerge.dist,
                SDSS_DR17_APOGEE_Allstarmerge.dist_src,
            )
            .join(CatalogToTIC_v8, on=(Catalog.catalogid == CatalogToTIC_v8.catalogid))
            .join(TIC_v8, on=(CatalogToTIC_v8.target_id == TIC_v8.id))
            .join(TwoMassPSC, on=(TIC_v8.twomass_psc == TwoMassPSC.designation))
            .join(
                SDSS_DR17_APOGEE_Allstarmerge,
                on=(
                    TwoMassPSC.designation
                    == peewee.fn.replace(SDSS_DR17_APOGEE_Allstarmerge.apogee_id, "2M", "")
                ),
            )
            .where(
                CatalogToTIC_v8.version_id == version_id,
                CatalogToTIC_v8.best >> True,
                *mwm_rv_long_condition,
                SDSS_DR17_APOGEE_Allstarmerge.h < 11.5,
            )
        )
        # Below ra, dec and radius are in degrees
        # query_region[0] is ra of center of the region
        # query_region[1] is dec of center of the region
        # query_region[2] is radius of the region
        if query_region:
            query = query.where(
                peewee.fn.q3c_radial_query(
                    Catalog.ra,
                    Catalog.dec,
                    query_region[0],
                    query_region[1],
                    query_region[2],
                )
            )
        return query

    def post_process(self, model):
        """
        For cadence:
        If H>10.8 then use bright_<nn>x2, otherwise use bright_<nn>x1,
        where <nn> = 3*ceiling((18-nvisits)/3)
        if <nn> is less than 6 then
            set <nn> = 6

        """

        # teff_avg and logg_avg are from SDSS_DR17_APOGEE_Allstarmerge
        # old name was teff, logg
        cursor = self.database.execute_sql(
            "select catalogid, nvisits, h, teff_avg, logg_avg from "
            + " sandbox.temp_mwm_bin_rv_long_apogee ;"
        )

        output = cursor.fetchall()

        for i in range(len(output)):
            current_catalogid = output[i][0]
            current_nvisits = output[i][1]
            current_h = output[i][2]
            # current_teff_avg = output[i][3]
            # current_logg_avg = output[i][4]

            nn = 3 * math.ceil((18 - current_nvisits) / 3)
            if nn < 6:
                nn = 6

            if current_h > 10.8:
                current_cadence = "bright_" + str(nn) + "x2"
            else:
                current_cadence = "bright_" + str(nn) + "x1"

            if current_cadence is not None:
                self.database.execute_sql(
                    " update sandbox.temp_mwm_bin_rv_long_apogee "
                    + " set cadence = '"
                    + current_cadence
                    + "'"
                    " where catalogid = " + str(current_catalogid) + ";"
                )


# Below is from the comments for
# MWM_bin_rv_short_Base_Carton further below.
# It is put here for reference.
#
# v1.0
# AND tm.j_msigcom <= 0.1
# AND tm.h_msigcom <= 0.1
# AND tm.k_msigcom <= 0.1
# AND (   tm.ph_qual='AAA'
#      OR tm.ph_qual='AAB'
#      OR tm.ph_qual='ABA'
#      OR tm.ph_qual='BAA'
#      OR tm.ph_qual='ABB'
#      OR tm.ph_qual='BAB'
#      OR tm.ph_qual='BBA'
#      OR tm.ph_qual='BBB')
# AND tm.prox >= 6
# AND tm.cc_flg='000'
# AND tm.gal_contam='000'
# AND  (   tm.rd_flg='111'
#       OR tm.rd_flg='112'
#       OR tm.rd_flg='121'
#       OR tm.rd_flg='211'
#       OR tm.rd_flg='122'
#       OR tm.rd_flg='212'
#       OR tm.rd_flg='221'
#       OR tm.rd_flg='222')
# AND g.parallax_error/g.parallax < 0.2

mwm_rv_short_condition = (
    TwoMassPSC.j_msigcom <= 0.1,
    TwoMassPSC.h_msigcom <= 0.1,
    TwoMassPSC.k_msigcom <= 0.1,
    (
        (TwoMassPSC.ph_qual == "AAA")
        | (TwoMassPSC.ph_qual == "AAB")
        | (TwoMassPSC.ph_qual == "ABA")
        | (TwoMassPSC.ph_qual == "BAA")
        | (TwoMassPSC.ph_qual == "ABB")
        | (TwoMassPSC.ph_qual == "BAB")
        | (TwoMassPSC.ph_qual == "BBA")
        | (TwoMassPSC.ph_qual == "BBB")
    ),
    TwoMassPSC.prox >= 6,
    TwoMassPSC.cc_flg == "000",
    TwoMassPSC.gal_contam == 0,
    (
        (TwoMassPSC.rd_flg == "111")
        | (TwoMassPSC.rd_flg == "112")
        | (TwoMassPSC.rd_flg == "121")
        | (TwoMassPSC.rd_flg == "211")
        | (TwoMassPSC.rd_flg == "122")
        | (TwoMassPSC.rd_flg == "212")
        | (TwoMassPSC.rd_flg == "221")
        | (TwoMassPSC.rd_flg == "222")
    ),
)


class MWM_bin_rv_short_Base_Carton(BaseCarton):
    # old class name MWM_RV_Short_FPS_Carton(BaseCarton):
    """
    This base carton contains the first part of the below query.
    which is common to all the mmw_bin_rv_short* cartons
    The middle part (i.e. the second part) is different for the different cartons.
    The third part is not in mwm_bin_rv_short_mdwarf.
    The third part is in mwm_bin_rv_short_subgiant and mwm_bin_rv_short_rgb.
    The third part corresponds to mwm_rv_short_condition above.

    The below information is for 5.1.25. mwm_bin_rv_short_mdwarf.
    It is put here for reference for MWM_bin_rv_short_Base_Carton.
    The below information is again put further below in 5.1.25. mwm_bin_rv_short_mdwarf.

    5.1.25. mwm_bin_rv_short_mdwarf
    This carton contains the M Dwarf stars originally in mwm_bin_rv_short

    Shorthand name:  mwm_bin_rv_short_mdwarf  (Formally mwm_rv_short_fps/mwm_bin_rv_short)

    Existing carton code :
    https://github.com/sdss/target_selection/blob/main/python/target_selection/cartons/mwm_rv.py

    Simplified Description of selection criteria :

    SELECT catalog.catalogid, catalog.ra, catalog.dec, catalog.pmra, catalog.pmdec,
    g.ref_epoch, g.source_id, g.parallax, g.parallax_error,
    g.phot_g_mean_mag,g.phot_bp_mean_mag,
    g.phot_rp_mean_mag,g.logg_gspphot,g.teff_gspphot,tm.j_m,tm.h_m,tm.k_m
    FROM catalog
    JOIN gaia_dr3_source AS g,twomass_psc AS tm

    WHERE tm.h_m < 11.5
    AND tm.h_m > 7
    AND g.parallax>0
    AND g.phot_rp_mean_mag-tm.k_m>2 and
    AND (g.phot_rp_mean_mag-tm.k_m)*2+0.6<tm.h_m-5*log10(1000/g.parallax)+5

    Gaia DR2 parameters to be converted to Gaia DR3 :
    coordinates, proper motions, parallax

    Return columns (again ok to copy from v0.5 version) :  (See pseudo-SQL above)

    Metadata:

    can_offset = False
    Cadence options: bright_18x1
    priority: priority = 2515
    Lead contact:  Nicholas Troup Nathan De Lee
    """

    # Below query does not use SDSS_DR17_APOGEE_Allstarmerge.
    # There is gaia_dr3 in the below query so we have done
    # major modification of the old v0.5 query.

    def build_query(self, version_id, query_region=None):
        query = (
            Catalog.select(
                Catalog.catalogid,
                Catalog.ra,
                Catalog.dec,
                Catalog.pmra,
                Catalog.pmdec,
                Gaia_DR3.ref_epoch,
                Gaia_DR3.source_id,
                Gaia_DR3.parallax,
                Gaia_DR3.parallax_error,
                Gaia_DR3.phot_g_mean_mag,
                Gaia_DR3.phot_bp_mean_mag,
                Gaia_DR3.phot_rp_mean_mag,
                Gaia_DR3.logg_gspphot,
                Gaia_DR3.teff_gspphot,
                TwoMassPSC.j_m,
                TwoMassPSC.h_m,
                TwoMassPSC.k_m,
                Gaia_DR3.ra.alias("gaia_dr3_ra"),
                Gaia_DR3.dec.alias("gaia_dr3_dec"),
                Gaia_DR3.pmra.alias("gaia_dr3_pmra"),
                Gaia_DR3.pmdec.alias("gaia_dr3_pmdec"),
            )
            .join(CatalogToGaia_DR3, on=(Catalog.catalogid == CatalogToGaia_DR3.catalogid))
            .join(Gaia_DR3, on=(CatalogToGaia_DR3.target_id == Gaia_DR3.source_id))
            .switch(Catalog)
            .join(CatalogToTwoMassPSC, on=(Catalog.catalogid == CatalogToTwoMassPSC.catalogid))
            .join(TwoMassPSC, on=(CatalogToTwoMassPSC.target_id == TwoMassPSC.pts_key))
            .where(
                CatalogToTwoMassPSC.version_id == version_id,
                CatalogToTwoMassPSC.best >> True,
                CatalogToGaia_DR3.version_id == version_id,
                CatalogToGaia_DR3.best >> True,
                TwoMassPSC.ext_key >> None,
                Gaia_DR3.parallax.is_null(False),
                Gaia_DR3.parallax > 0,
            )
        )

        # Below ra, dec and radius are in degrees
        # query_region[0] is ra of center of the region
        # query_region[1] is dec of center of the region
        # query_region[2] is radius of the region
        if query_region:
            query = query.where(
                peewee.fn.q3c_radial_query(
                    Catalog.ra,
                    Catalog.dec,
                    query_region[0],
                    query_region[1],
                    query_region[2],
                )
            )
        return query


class MWM_bin_rv_short_mdwarf_Base_Carton(MWM_bin_rv_short_Base_Carton):
    """5.1.25. mwm_bin_rv_short_mdwarf
    This base carton contains the M Dwarf stars originally in mwm_bin_rv_short

    Shorthand name:  mwm_bin_rv_short_mdwarf  (Formally mwm_rv_short_fps/mwm_bin_rv_short)

    Existing carton code :
    https://github.com/sdss/target_selection/blob/main/python/target_selection/cartons/mwm_rv.py

    Simplified Description of selection criteria :

    SELECT catalog.catalogid, catalog.ra, catalog.dec, catalog.pmra, catalog.pmdec,
    g.ref_epoch, g.source_id, g.parallax, g.parallax_error,
    g.phot_g_mean_mag,g.phot_bp_mean_mag,
    g.phot_rp_mean_mag,g.logg_gspphot,g.teff_gspphot,tm.j_m,tm.h_m,tm.k_m
    FROM catalog
    JOIN gaia_dr3_source AS g,twomass_psc AS tm

    WHERE tm.h_m < 11.5
    AND tm.h_m > 7
    AND g.parallax>0
    AND g.phot_rp_mean_mag-tm.k_m>2 and
    AND (g.phot_rp_mean_mag-tm.k_m)*2+0.6<tm.h_m-5*log10(1000/g.parallax)+5

    The mwm_bin_rv_short_mdwarf carton has the above query with only two parts.
    The mwm_bin_rv_short_subgiant and mwm_bin_rv_short_rgb cartons
    have queries with three parts
    The third part corresponds to mwm_rv_short_condition above.

    Gaia DR2 parameters to be converted to Gaia DR3 :
    coordinates, proper motions, parallax

    Return columns (again ok to copy from v0.5 version) : (See pseudo-SQL above)

    Metadata:

    can_offset = False
    Cadence options: bright_18x1
    priority: priority = 2515
    Lead contact:  Nicholas Troup Nathan De Lee
    """

    def build_query(self, version_id, query_region=None):
        query = super().build_query(version_id, query_region)
        query = query.where(
            TwoMassPSC.h_m > 7,
            TwoMassPSC.h_m < 11.5,
            Gaia_DR3.phot_rp_mean_mag - TwoMassPSC.k_m > 2,
            ((Gaia_DR3.phot_rp_mean_mag - TwoMassPSC.k_m) * 2 + 0.6)
            < (TwoMassPSC.h_m - 5 * peewee.fn.log10(1000 / Gaia_DR3.parallax) + 5),
        )
        return query


class MWM_bin_rv_short_mdwarf_apogee_18epoch(MWM_bin_rv_short_mdwarf_Base_Carton):
    name = "mwm_bin_rv_short_mdwarf_apogee_18epoch"
    category = "science"
    instrument = "APOGEE"
    cadence = "bright_18x1"
    program = "mwm_bin"
    mapper = "MWM"
    priority = 1305
    can_offset = False


class MWM_bin_rv_short_mdwarf_apogee_12epoch(MWM_bin_rv_short_mdwarf_Base_Carton):
    name = "mwm_bin_rv_short_mdwarf_apogee_12epoch"
    category = "science"
    instrument = "APOGEE"
    cadence = "bright_12x1"
    program = "mwm_bin"
    mapper = "MWM"
    priority = 1306
    can_offset = False


class MWM_bin_rv_short_mdwarf_apogee_08epoch(MWM_bin_rv_short_mdwarf_Base_Carton):
    name = "mwm_bin_rv_short_mdwarf_apogee_08epoch"
    category = "science"
    instrument = "APOGEE"
    cadence = "bright_8x1"
    program = "mwm_bin"
    mapper = "MWM"
    priority = 1307
    can_offset = False


class MWM_bin_rv_short_subgiant_apogee_Carton(MWM_bin_rv_short_Base_Carton):
    """5.1.26. mwm_bin_rv_short_subgiant_apogee
    This carton contains subgiant stars and lower red giant branch stars
    originally in mwm_bin_rv_short

    Shorthand name:  mwm_bin_rv_short_subgiant  (Formally mwm_rv_short_fps/mwm_bin_rv_short)

    Existing carton code :
    https://github.com/sdss/target_selection/blob/main/python/target_selection/cartons/mwm_rv.py

    Simplified Description of selection criteria :

    SELECT catalog.catalogid, catalog.ra, catalog.dec, catalog.pmra, catalog.pmdec,
    g.ref_epoch, g.source_id, g.parallax, g.parallax_error,
    g.phot_g_mean_mag,g.phot_bp_mean_mag,
    g.phot_rp_mean_mag,g.logg_gspphot,g.teff_gspphot,tm.j_m,tm.h_m,tm.k_m
    FROM catalog
    JOIN gaia_dr3_source AS g,twomass_psc AS tm

    WHERE tm.h_m <10.8
    AND tm.h_m > 7
    AND g.logg_gspphot > 3.0
    AND g.logg_gspphot < 4.1
    AND g.teff_gspphot < 7000
    AND g.teff_gspphot > 4500

    AND (tm.j_msigcom <= 0.1 AND tm.h_msigcom<=0.1 AND tm.k_msigcom <= 0.1)
    AND(tm.ph_qual= 'AAA' OR tm.ph_qual='AAB' OR tm.ph_qual='ABA'
        OR tm.ph_qual='BAA' OR tm.ph_qual= 'ABB' OR tm.ph_qual='BAB'
        OR tm.ph_qual='BBA' OR tm.ph_qual ='BBB')
    AND tm.prox >= 6  AND tm.cc_flg='000' AND tm.gal_contam='000'
    AND  (tm.rd_flg='111' OR tm.rd_flg='112'  OR tm.rd_flg='121'
          OR tm.rd_flg='211'   OR tm.rd_flg='122'  OR tm.rd_flg='212'
          OR tm.rd_flg='221'  OR tm.rd_flg='222')
    AND g.parallax_error/g.parallax < 0.2

    Gaia DR2 parameters to be converted to Gaia DR3 :
    coordinates, proper motions, parallax

    Return columns (again ok to copy from v0.5 version) : (See pseudo-SQL above)

    Metadata:

    can_offset = False
    Cadence options: bright_18x1
    priority: priority = 2525
    Lead contact:  Nicholas Troup Nathan De Lee
    """

    name = "mwm_bin_rv_short_subgiant_apogee"
    category = "science"
    instrument = "APOGEE"
    cadence = "bright_18x1"
    program = "mwm_bin"
    mapper = "MWM"
    priority = 2525
    can_offset = False

    def build_query(self, version_id, query_region=None):
        query = super().build_query(version_id, query_region)
        query = query.where(
            TwoMassPSC.h_m > 7,
            TwoMassPSC.h_m < 10.8,
            Gaia_DR3.logg_gspphot > 3.0,
            Gaia_DR3.logg_gspphot < 4.1,
            Gaia_DR3.teff_gspphot < 7000,
            Gaia_DR3.teff_gspphot > 4500,
            Gaia_DR3.parallax_error / Gaia_DR3.parallax < 0.2,
            *mwm_rv_short_condition,
        )
        return query


class MWM_bin_rv_short_rgb_apogee_Carton(MWM_bin_rv_short_Base_Carton):
    """5.1.27. mwm_bin_rv_short_rgb_apogee
    This carton contains the red clump and higher red giant stars
    originally in mwm_bin_rv_short

    Shorthand name:  mwm_bin_rv_short_rgb  (Formally mwm_rv_short_fps/mwm_bin_rv_short)

    Existing carton code :
    https://github.com/sdss/target_selection/blob/main/python/target_selection/cartons/mwm_rv.py

    Simplified Description of selection criteria :

    SELECT catalog.catalogid, catalog.ra, catalog.dec, catalog.pmra, catalog.pmdec,
    g.ref_epoch, g.source_id, g.parallax, g.parallax_error,
    g.phot_g_mean_mag,g.phot_bp_mean_mag,
    g.phot_rp_mean_mag,g.logg_gspphot,g.teff_gspphot,tm.j_m,tm.h_m,tm.k_m
    FROM catalog
    JOIN gaia_dr3_source AS g,twomass_psc AS tm

    WHERE tm.h_m<10.8
    AND tm.h_m > 7
    AND g.logg_gspphot > 1.2
    AND g.logg_gspphot <= 3.0
    AND g.teff_gspphot < 5500

    AND (tm.j_msigcom <= 0.1 AND tm.h_msigcom<=0.1 AND tm.k_msigcom <= 0.1)
    AND(tm.ph_qual= 'AAA' OR tm.ph_qual='AAB' OR tm.ph_qual='ABA'
        OR tm.ph_qual='BAA' OR tm.ph_qual= 'ABB' OR tm.ph_qual='BAB'
        OR tm.ph_qual='BBA' OR tm.ph_qual ='BBB')
    AND tm.prox >= 6  AND tm.cc_flg='000' AND tm.gal_contam='000'
    AND  (tm.rd_flg='111' OR tm.rd_flg='112'  OR tm.rd_flg='121'
          OR tm.rd_flg='211'   OR tm.rd_flg='122'  OR tm.rd_flg='212'
          OR tm.rd_flg='221'  OR tm.rd_flg='222')
    AND g.parallax_error/g.parallax < 0.2

    Gaia DR2 parameters to be converted to Gaia DR3 :
    coordinates, proper motions, parallax

    Return columns (again ok to copy from v0.5 version) : (See pseudo-SQL above)

    Metadata:

    can_offset = False
    Cadence options: bright_18x1
    priority: priority = 2535
    Lead contact:  Nicholas Troup Nathan De Lee
    """

    name = "mwm_bin_rv_short_rgb_apogee"
    category = "science"
    instrument = "APOGEE"
    cadence = "bright_18x1"
    program = "mwm_bin"
    mapper = "MWM"
    priority = 2535
    can_offset = False

    def build_query(self, version_id, query_region=None):
        query = super().build_query(version_id, query_region)
        query = query.where(
            TwoMassPSC.h_m > 7,
            TwoMassPSC.h_m < 10.8,
            Gaia_DR3.logg_gspphot > 1.2,
            Gaia_DR3.logg_gspphot <= 3.0,
            Gaia_DR3.teff_gspphot < 5500,
            Gaia_DR3.parallax_error / Gaia_DR3.parallax < 0.2,
            *mwm_rv_short_condition,
        )
        return query


class MWM_bin_rv_short_rgb_apogee_08epoch_Carton(MWM_bin_rv_short_rgb_apogee_Carton):
    """mwm_bin_rv_short_rgb_apogee_08epoch -
    same as mwm_bin_rv_short_rgb_apogee but
    with cadence=bright_8x1 and priority=2537
    """

    name = "mwm_bin_rv_short_rgb_apogee_08epoch"
    category = "science"
    instrument = "APOGEE"
    cadence = "bright_8x1"
    program = "mwm_bin"
    mapper = "MWM"
    priority = 2537
    can_offset = True


class MWM_bin_rv_short_rgb_apogee_12epoch_Carton(MWM_bin_rv_short_rgb_apogee_Carton):
    """mwm_bin_rv_short_rgb_apogee_12epoch -
    same as mwm_bin_rv_short_rgb_apogee but
    with cadence=bright_12x1 and priority=2536
    """

    name = "mwm_bin_rv_short_rgb_apogee_12epoch"
    category = "science"
    instrument = "APOGEE"
    cadence = "bright_12x1"
    program = "mwm_bin"
    mapper = "MWM"
    priority = 2536
    can_offset = True


class MWM_bin_rv_short_subgiant_apogee_08epoch_Carton(MWM_bin_rv_short_subgiant_apogee_Carton):
    """mwm_bin_rv_short_subgiant_apogee_08epoch -
    same as mwm_bin_rv_short_subgiant_apogee but
    with cadence=bright_8x1 and priority=2527
    """

    name = "mwm_bin_rv_short_subgiant_apogee_08epoch"
    category = "science"
    instrument = "APOGEE"
    cadence = "bright_8x1"
    program = "mwm_bin"
    mapper = "MWM"
    priority = 2527
    can_offset = True


class MWM_bin_rv_short_subgiant_apogee_12epoch_Carton(MWM_bin_rv_short_subgiant_apogee_Carton):
    """mwm_bin_rv_short_subgiant_apogee_12epoch -
    same as mwm_bin_rv_short_subgiant_apogee but
    with cadence=bright_12x1 and priority=2526
    """

    name = "mwm_bin_rv_short_subgiant_apogee_12epoch"
    category = "science"
    instrument = "APOGEE"
    cadence = "bright_12x1"
    program = "mwm_bin"
    mapper = "MWM"
    priority = 2526
    can_offset = True



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: Pramod Gupta (psgupta@uw.edu)
# @Date: 2023-05-22
# @Filename: mwm_cb_cvcandidates.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

import peewee

from sdssdb.peewee.sdss5db.catalogdb import CataclysmicVariables, Catalog, CatalogToTIC_v8, TIC_v8

from target_selection.cartons import BaseCarton


class MWM_CB_CV_Candidates_Carton(BaseCarton):
    """
    Definition:
        List of cataclysmic variables (+candidates)
        compiled by the AAVSO.
    SQL:
        Select all records from table cataclysmic_variables.
    Cadence: This target sample will be split into
        (1) H<11 to be observed with APOGEE.
        (2) phot_g_mean_mag < 16 to be observed with BOSS in bright time.
        (3) phot_g_mean_mag > 16 to be observed with BOSS in dark time.
    This is a base carton. Actual cartons are implemented as subclasses for the
    different magnitude cuts.
    """

    def build_query(self, version_id, query_region=None):
        # The column CataclysmicVariables.source_id
        # corresponds to Gaia_DR2 source_id.
        query = (
            CatalogToTIC_v8.select(
                CatalogToTIC_v8.catalogid,
                CataclysmicVariables.source_id,
                CataclysmicVariables.phot_g_mean_mag,
                TIC_v8.hmag.alias("h"),
            )
            .join(TIC_v8)
            .join(CataclysmicVariables, on=(CataclysmicVariables.source_id == TIC_v8.gaia_int))
            .where(CatalogToTIC_v8.version_id == version_id, CatalogToTIC_v8.best >> True)
        )

        if query_region:
            query = query.join_from(CatalogToTIC_v8, Catalog).where(
                peewee.fn.q3c_radial_query(
                    Catalog.ra,
                    Catalog.dec,
                    query_region[0],
                    query_region[1],
                    query_region[2],
                )
            )

        return query


class MWM_CB_CV_Candidates_APOGEE_Carton(MWM_CB_CV_Candidates_Carton):
    """MWM CB CV to be observed with APOGEE."""

    name = "mwm_cb_cvcandidates_apogee"
    mapper = "MWM"
    category = "science"
    program = "mwm_cb"
    instrument = "APOGEE"
    cadence = "bright_1x1"
    priority = 1820
    can_offset = True

    def build_query(self, version_id, query_region=None):
        query = super().build_query(version_id, query_region=query_region)

        return query.where(TIC_v8.hmag < 11)


class MWM_CB_CV_Candidates_BOSS_Carton(MWM_CB_CV_Candidates_Carton):
    """MWM CB CV to be observed with BOSS."""

    name = "mwm_cb_cvcandidates_boss"
    mapper = "MWM"
    category = "science"
    program = "mwm_cb"
    instrument = "BOSS"
    cadence = None  # cadence is set in post_process()
    priority = 1820
    can_offset = True

    def post_process(self, model, **kwargs):
        # G > 16 => cadence = dark_2x1
        model.update(cadence="dark_2x1").where(model.phot_g_mean_mag > 16).execute()

        # G < 16 => cadence = bright_2x1
        model.update(cadence="bright_2x1").where(model.phot_g_mean_mag < 16).execute()

        return model



from .append_to_sdss_id import (
    AppendToTables,
    SdssIdFlat,
    SdssIdFlatAddendum,
    SdssIdStacked,
    SdssIdStackedAddendum,
    TempCatalogidV21,
    TempCatalogidV25,
    TempCatalogidV31,
)
from .create_catalogidx_to_catalogidy import (
    MetaXMatch,
    TempMatch,
    UniqueMatch,
    create_unique_from_region,
)
